#!/usr/bin/env python3
#@ S-Postman is an unsophisticated mail fetcher/processor/sender/archiver,
#@ ment to be used by normal end-users behind an administrated account.
# FIXME:
#   - use GLOBAL-FLAG  mode  MODE-FLAG command line scheme
#   - offer os.path.expandvars for RC values
#   - also offer datetime().strftime() support for RC's
#   - send mode should at least offer the stuff fetchmail(1) uses to rewrite
#     from/to etc.
#   - IMAP support!! imap-action= move=,delete; imap-folders = inbox,fast-fetch...
#     I.e., add a [folder] directive which works exactly as [box] in respect to
#     PLUG and friends (except not local but folder MOVE/COPY??? on server)
# FIXME --archify and --awake - move entire mailbox to an .mbox and compress
# that via [box] mummy = zip|gzip|bzip2
# these are monthly archives?!  optionally? create place a single mail in the
# now empty source mailbox which contains only the plain From/To/Date/Subject
# fields of all archived mails (Subject: Archiev YYYY-MM-NAME)
# what abot wekkly as in YYYY-WW??
# shouldn't this be an automatic buitin, i.e. all [box]es wirh an mummy option
# get mummied via --mummify??  i.e. NOT like dispatch
#
# Created: 2011-01-10
BUGMAIL = ('All posts to the https://bitbucket.org/sdaoden/s-postman '
           'issue tracker, please')
__version__ = 'S-Postman v0.4.0a'
__copyright__ = """
Copyright (c) 2011 Steffen Daode Nurpmeso <sdaoden@gmail.com>.
All rights reserved.
This software is published under the terms of the BSD 2 license.
"""
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the author nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import atexit, os, re, sys, traceback
import email, getopt, getpass, mailbox, re, stat, time

# For non-Py3K this fails with SyntaxError and shows message, just as desired.
if sys.hexversion < 0x030200C2:
    print('This script requires Python 3000, version 3.2rc2 or above.',
          file=sys.stderr)
    sys.exit(1)

# RST {{{
RST = """.. -*-restructuredtext-*-

************
s-postman.py
************

==========================
An end-user mail processor
==========================

:Manual section: 1
:Manual group:   S-Postman Manual
:Version:        0.4.0a
:Date:           2011-03-11

.. |s-p(1)| replace:: *s-postman*\ (1)

""" # }}}

# USAGE {{{
USAGE = """
Usage
=====

    -h, --help  print usage information and exit
    -d, --doc   print rST documentation (long output) and exit.
                If you have docutils installed, you may use one of the various
                ``rst2XY`` programs to produce a nicely formatted version of it
    -v, --version
                print version information and exit
    -V, --verbose
                be verbose, given twice even more
    -m DIR, --maildir=DIR, --md=DIR
                specify that output mail ``DIR``\ ectory is to be used
    -r FILE, --resource=FILE, --rc=FILE
                specify the resource ``FILE`` to be used.
                If this option is missing the ``S_POSTMAN_RC`` environment
                variable is tried, falling back ``~/.s-postmanrc`` as
                a last option
    -I, --ignore-errors
                continue some non-fatal errors without prompting the user
    -C, --check-resource
                load and process resource file etc. and exit.  Useful in
                conjunction with ``-VV`` to test a setup
    -F ACCOUNT, --fetch=ACCOUNT
                fetch new mail for the ``[account]`` ``ACCOUNT``, if available.
    -A, --fetchall
                fetch new mail for all configured ``[account]``\ s.
                This is the default mode of operation
    -L, --fetchloop
                enter a loop which periodically performs a --fetchall option.
                (You can use the job control of your POSIX compliant shell
                to place S-Postman in the background)
    -D MAILBOX, --dispatch=MAILBOX
                open ``MAILBOX`` and process it with the configured rules.
                The mailbox type defaults to ``mbox``, but it is possible to
                give this parameter as a ``MAILBOX:TYPE`` tuple, e.g.
                ``--dispatch=somefile:maildir`` would treat ``somefile`` as
                a mailbox in ``maildir`` format - please see --doc for all
                available mailbox types.
                It is undefined what happens if ``MAILBOX`` may itself
                match as a target for any ``[rule]``.
                E.g. a *UNIX* user may do ``s-postman.py -D $MAIL`` to
                process it's local mail
    -k, --keep-archives
                don't treat digest (archive) mails as such, even if
                a ``[rule]`` matches which is configured to split them up
    -s START, --start=START
                Better support for restarting after a crash (ooops).
                For dispatching this option can be used to specify a starting
                ticket number (e.g. if a crash happened after processing 20
                mails, say ``--start=21``), for fetching this specifies the
                temporary download folder (where yet downloaded but sofar
                unprocessed data is stored)

""" # }}}

# __doc__ {{{
__doc__ = """
Description
===========

|s-p(1)| is a `python`_ tool which is capable to fetch and dispatch mail.
It is ment for end-users:
this means that it's fetching capabilities do not reach those of
`fetchmail`_, nor do it's dispatching capabilities reach those of `procmail`_,
but it's easy to configure, has a quite powerful rule-matching syntax, and,
because it is a single monolithic program, it may even do it's work much
faster than the combination of the mentioned.

..  _python: http://www.python.org
..  _fetchmail: http://fetchmail.berlios.de
..  _procmail: http://www.procmail.org


Configuration
=============

Because the resource configuration file may contain account credentials
|s-p(1)| will panic if any other user but the owner can read that file.


Syntax of configuration files
-----------------------------

Configuration files are treated as a sequence of lines.
Lines starting with ``#`` are comments.
Comments and empty lines are discarded.
Lines consisting of ``[section-name]`` create/switch to a configuration section
named *section-name*.
Other lines are expected to consist of ``key = value`` pairs;
note that whitespace after *key* and before *value* is discarded.
*value* lines can be continued over as many lines as required if the newlines
are "escaped" with a backslash, as in::

    some-key = this value line \\
                spreads over \\
                multiple lines.

All leading whitespace of follow lines is discarded, but *not* any trailing
whitespace before the backslash.


Configuration directives
------------------------

The following example configuration also documents all supported directives::

    # [default] is active by default, but it can always be re-switched to and
    # continued by explicitely stating [default] on a line of it's own.
    [default]
    # Overwritten by --maildir
    # If neither --maildir no this option is set, the S_POSTMAN_MAILDIR
    # environment variable is inspected, and the last fallback is ~/mail
    maildir = absolute-maildir-path
    # Optional (defaults to 'mbox')
    mailbox-type = mbox|Maildir|Babyl|MH|MMDF
    # Optional
    # If this is given (and enabled) then the [account] entries don't need
    # to specify any of 'user' and 'pass' (so that malicious administrators
    # can't figure them out).
    # All operation modes which need [account] data (e.g. --fetch, --send)
    # will instead prompt on standart input for (missing) user credentials.
    ask-user = BOOLEAN (yes|true|ok|on and abbrevs thereof - else false)

    # Optional
    # If set, mail is piped into STDIN of a hook and read again from the hook's
    # STDOUT; note that either of those is optional, it is used what's set
    # FIXME about exit status handling
    fetch-hook = PROGRAM
    fetch-hook-largemail = PROGRAM
    # Optional (defaults to 500000)
    # How large (in bytes) a mail has to be to switch forth to
    # fetch-hook-largemail
    fetch-hook-treshold = BYTES

    # Optional (default to 1200 and 300, respectively)
    # In --fetchloop mode this specifies the time in seconds to wait before
    # looking for new mail.  The switch back from -fast is applied in steps
    loop-check = SECONDS
    loop-check-fast = SECONDS
    # Optional (defaults to 1)
    # How many mails are needed to switch forth to loop-check-fast
    loop-treshold = NUMBER
    # Optional
    # If this is set it is treated as a command (and arguments) to be run
    # (by the shell) if new mail has been downloaded and dispatched
    loop-hook-newmail = mpg321 path_to_mail_arrived_sound.mp3

    [account]
    id = unique identity
    account = user@somewhere.com
    # Optional account aliases (may be given multiple times)
    alias = somealias
    # User credentials (also see 'ask-user' directive in [default]).
    # pass is optional from S-Postman's point of view unless an encrypted
    # network protocol is used
    user = user-id-at-somewhere
    pass = user-pass-at-somwhere
    # Optional (defaults to "SMTP")
    send-type = smtp|smtps
    send-server = protocol.example.com
    # Optional (defaults to "POP")
    fetch-type = pop|pops
    fetch-server = protocol.example.com

    # Target, I: a local mailbox
    # Note: a mailbox with id NULL always exists - it's a data sink
    [box]
    id = unique name (or path)
    # Optional ("id" is used if this is not set)
    # If set user home directory expansion is performed
    # (maildir can thus be left if this is used...)
    path = path
    # Optional (defaults to global "mailbox-type")
    type = same as [default] mailbox-type

    IMAP # Target, II: an IMAP folder (located on an IMAP server)
    IMAP # This will apply to whatever [account] is currently processed
    IMAP [folder]
    IMAP id = unique name
    IMAP path = directory name on IMAP server

    # The ruleset is the dispatch work pipeline.  All mails are piped through
    # it, [rule]s are applied in the order in which they are specified here.
    # [ruleset] is single instance, but it can be continued by re-opening
    # the [ruleset] section as often as necessary.
    [ruleset]
    id-of-a-[rule] = id-of-a-[TARGET]
    id-of-another-[rule] = id-of-a-[TARGET], id-of-another-[TARGET], ...
    # All mails which reach this point without being consumed by a [rule]
    # will be stored in the *very* last [TARGET] which has been defined

    # Rules combine a series of commands to be applied against mails
    # Note: a rule with id NULL always exists - it's an all-match data sink;
    # if you use it in [ruleset] you don't need to specify any [TARGET]
    # (rule NULL is very handy as the last entry of [ruleset]...)
    [rule]
    id = unique name
    COMMANDS
    # .. COMMANDS are described in the section `Command syntax`_ below

    # The next line includes the file ".fingerprints", which is to looked up
    # *relative* to the directory where .s-postmanrc has been loaded from.
    # "~" user home directory is expanded, too.
    # Recursive inclusions are supported.
    # Access checks just as for the configuration file itself are not
    # performed, but it's a fatal error if FILE doesn't exist.
    # Note that %include finishs the currently active [section]
    %include .fingerprints

    # [hostfingerprints] is single instance, but it can be continued by
    # re-opening the [hostfingerprints] section as often as necessary
    # (May be in .fingerprints)
    [hostfingerprints]
    # FINGERPRINT is the fingerprint of the server's certificate.
    # Get the certificate:
    # openssl s_client -certform DER -showcerts -connect SERVER[:PORT]
    # Calculate fingerprint
    # FIXME  openssl x509 -fingerprint -sha1 -noout -in 
    HOSTNAME = FINGERPRINT


Command syntax
--------------

There are six kind of commands.
Where a ``regular expression`` is required a valid Python regular expression
is required; do ``pydoc3 re`` on your command line for an overview.

    *PASS*:     ``?header-field,header-field = regular expression``
        Break off this rule if the regular expression does *not* match.
    *HIT*:      ``!header-field,header-field = regular expression``
        End rule processing and save mail in all active ``[TARGET]``\ s
        if the regular expression matches.
    *BRANCH*:   ``@id-of-a-[rule][@header-field,..] = [regular expression]``
        Branch off to the specified ``rule`` if the regular expression matches.
        Note that both ``@header-field(s)`` and ``regular expression`` are
        optional - if only ``@id-of-a-[rule]`` is given then the branch is
        performed unconditionally.
        When branching off to another rule the active list of ``[TARGET]``\ s
        is copied, but no other state.  When the branch returns because the
        mail has not been consumed inside of it, then the state is restored to
        exactly the same state it was in before the branch has been entered.
        It is verified that nesting of rules does not exceed 99 levels;
        that is so to avoid endless loops in rule branching.
    *ANY*:      ``$ = value-ignored``
        End rule processing and save mail in all active ``[TARGET]``\ s.
        If this is used it must be the last rule, and it is only valid if at
        least one *PASS*, *HIT* or *BRANCH* has been seen (in the rule).
    *SPLITTER*: ``* = name-of-splitter (or no value)``
        Change the rule's idea of what splitter is used to split archive mails.
        This defaults to no value at all meaning that no splitter is used.
        Supported mail-archive splitters:
        - text-digest
        FIXME - mime-digest
    *PLUG*:     ``: = id-of-a-[TARGET] (or no value)``
        Change the rules' list of active targets to
        ``id-of-a-defined-[TARGET]``.
        Without a value it resets the targets to the list which was active at
        entry of the ``[rule]``.
    *PLUG-ADD*: ``:+ = id-of-a-[TARGET]``
        Add a target to the current list of targets.  ``NULL`` is not allowed.
    *PLUG-SUB*: ``:- = id-of-a-[TARGET]``
        Remove a target from the current list of targets.  ``NULL`` is not
        allowed.
        *Note:* it causes a runtime error log message if the specified target
        is not part of the current list of targets.

E.g., the following hypothetic ``[ruleset]``/``[rule]`` combination does the
following:

    1.  The ruleset will start to match mail against the rule ``SomeDomain``.
    2.  The rule lets mail *PASS* which comes ``From:`` ``@SomeDomain.org``,
        goes ``To:`` this server or which is carbon-copied (``Cc:``) to there.
        If this does not match the rule returns processing to the ruleset.
    3.  The rule specifies that all mail from the ``security-announce``
        mailing-list is a *HIT* and thus is saved in ``some-domain``.
        If this command matches then the mail is consumed.
    4.  *PLUG* changes the list of targets to ``mailbox-id-two`` (only).
    5.  *SPLITTER* defines an archive splitter to be used from now on.
    6.  If the ``Subject:`` of a mail contains the string ``digest,``,
        then pipe it's contents through the ``text-digest`` splitter and
        save all splitted mails.
        If this command matches then the mail is consumed.
    7.  The rule is completed.  If the mail reaches this point without being
        consumed then processing control is passed back to the ruleset.
    8.  Because only a single rule is defined in the ruleset, all non-consumed
        mail will be saved in the last ``[TARGET]`` defined.

::

    [ruleset]
    SomeDomain = mailbox-id-one

    [rule]
    id = DomeDomain
    ?from,to,cc = @SomeDomain\.org
    !to,cc = (?:owner-)?security-announce
    : = mailbox-id-two
    * = text-digest
    !subject = digest,

Note that ``subject``, ``from``, ``to``, ``cc`` and ``bcc`` are treated special
and may thus be abbreviated as ``s,f,t,c,b``, respectively.


TODO:
=====

- Add POP protocol uidl/keep support.
  For now all mails are re-fetched over and over again unless the POP3 server
  deletes all downloaded mails (as it should, since ``keep`` is not set).
- Add IMAP support.  The source is somewhat prepared for that, but ...
- Extreme fail-safety is necessary at latest then.  Transaction log etc..
- Once Python EMAIL6 is released it may be useful to use file-based backend.
  Yet *all* mails are loaded into memory and held there until the final flush.
""" # }}}

VERBOSE = 0
POSTMAN_RC = '~/.s-postmanrc'
POSTMAN_MAILDIR = '~/mail'

ESTAT = 0

# Support funs (which should be in a module for sure) # {{{

# Logging stuff
def log(*args):
    if not VERBOSE:
        return
    print('| ' + ''.join(str(i) for i in args), file=sys.stderr)

def verb(*args):
    if VERBOSE < 2:
        return
    print('# ' + ''.join(str(i) for i in args), file=sys.stderr)

# Error handling
# - Due to usage of atexit do not reraise but sys.exit()!
# Shorten use of extb=E() of panic()/error()
E = sys.exc_info

def _print_extb(extb):
    print('______', file=sys.stderr)
    # Do an inline traceback.format_list() to avoid later replace()s
    tb = traceback.extract_tb(extb[2])
    if VERBOSE:
        print('Traceback (most recent call last):', file=sys.stderr)
    else:
        print('Traceback (raising call only):', file=sys.stderr)
        tb = [tb[-1]]

    for fname, lno, fun, line in tb:
        item = '  File "' + fname + '", line ' + str(lno) + ', in ' + fun
        if line:
            item += '\n    ' + line.strip()
        print(item, file=sys.stderr)

    for exfmt in traceback.format_exception_only(extb[0], extb[1]):
        print('Exception: ', exfmt, sep='', end='', file=sys.stderr)
    print('______', file=sys.stderr)

def panic(*args, extb=None):
    if extb:
        _print_extb(extb)
    print('PANIC: ' + ''.join(str(i) for i in args).replace('\n', '\n       '),
          file=sys.stderr)
    # FIXME print(BUGMAIL, file=sys.stderr)
    sys.exit(1)

def error(*args, logonly=False, extb=None):
    if extb:
        _print_extb(extb)
    print('ERROR: ' + ''.join(str(i) for i in args).replace('\n', '\n       '),
          file=sys.stderr)
    if logonly:
        return

    global ESTAT
    ESTAT = 1
    if Config.main['ignore-errors']:
        log('     --ignore-errors set, continuing')
        return
    if not user_confirm():
        print('Exit due to errors as above', file=sys.stderr)
        sys.exit(1)

def user_confirm(question='Continue [yY - else no]?'):
    try:
        b = input(question + ' ')
    except KeyboardInterrupt:
        panic('keyboard interrupt, terminating')
    except:
        b = 'n'
    return bool_check(b)

def user_line(intro='Enter line, confirm with ENTER:'):
    try:
        l = input(intro + ' ')
    except KeyboardInterrupt:
        panic('keyboard interrupt, terminating')
    except:
        l = ''
    return l

def user_pass(intro='Enter password, confirm with ENTER:'):
    base_intro = intro
    confirmed = None
    while 1:
        try:
            l = getpass.getpass(intro + ' ')
        except (KeyboardInterrupt, EOFError):
            panic('keyboard interrupt, terminating')
        except getpass.GetPassWarning:
            pass
        except:
            l = ''
        if not confirmed:
            confirmed = l
            intro = 'Please retype password: '
        elif confirmed != l:
            print('Password mismatch, restarting', file=sys.stderr)
            confirmed = None
            intro = base_intro
        else:
            break
    return confirmed

__bool_re = re.compile(r'\s*y(?:e(?:s)?)?|t(?:r(?:u(?:e)?)?)?|o(?:k|n)?', re.I)
def bool_check(val):
    return not not __bool_re.match(str(val))
# }}}

class Splitter: # {{{
    """Encapsulator for mail-archive splitters.
    Archive splitters (``msg`` argument is a ``email.message.Message``,
    their return type is a list of splitted messages or a list consisting of
    only the input message upon failure.
    """

    @staticmethod
    def create_if_supported(name):
        s = None
        name = name.lower()
        if name == 'text-digest':
            s = Splitter._text_digest
        elif name == 'mime-digest':
            s = _mime_digest
        return (name, s)

    @staticmethod
    def _text_digest(msg):
        """Text digest splitter"""
        # Real splitter: nl, MSG_SPLITTER, nl, (Message:.. nl)?, Date: header..
        START_SPLITTER = (b'------------------------------' +
                           b'----------------------------------------')
        MSG_SPLITTER = b'------------------------------'

        def __create_msg(lines):
            try:
                lastnl = False
                fp = email.feedparser.BytesFeedParser()
                while len(lines) > 0:
                    l = lines.pop(0)
                    if l.strip() == MSG_SPLITTER and lastnl:
                        break
                    lastnl = not len(l.strip())
                    fp.feed(l + b'\n')
                return fp.close()
            except Exception as e:
                log('    @ Error occurred - not splitting: ', str(e))
                return None

        if msg.is_multipart():
            log('    @ Multipart message: not splitting')
            return None
        lines = msg.get_payload(decode=True).splitlines()

        # Skip TOC first
        while len(lines):
            l = lines.pop(0)
            if l == START_SPLITTER:
                break
        result = []
        while len(lines):
            l = lines[0]
            if (l.startswith(b'Message: ') and len(lines) > 0 and
                lines[1].startswith(b'Date: ')):
                lines.pop(0)
                l = lines[0]
            if l.startswith(b'Date: '):
                nm = __create_msg(lines)
                if not nm:
                    return None
                result.append(nm)
            else:
                lines.pop(0)
        return result

    @staticmethod
    def _mime_digest(msg):
        """MIME-digest splitter"""
        #if msg.is_multipart():
        #    log('      @ Multipart message: not splitting')
        #    return [msg]
        return None
# }}}

# "Basic" S-Postman implementation follows #

class MailboxHelper: # {{{
    """Some functions which deal with ``mailbox.*``."""
    @staticmethod
    def name_if_supported(type):
        type = type.lower()
        if (type == 'mbox' or type == 'maildir' or type == 'babyl' or
                type == 'mh' or type == 'mmdf'):
            return type
        return None

    @staticmethod
    def open(path, type='mbox', create=False):
        log('Opening ', type, ' mailbox ', path)
        if type == 'mbox':
            mb = mailbox.mbox
        elif type == 'maildir':
            mb = mailbox.Maildir
        elif type == 'babyl':
            mb = mailbox.Babyl
        elif type == 'mh':
            mb = mailbox.MH
        else:
            mb = mailbox.MMDF
        try:
            mb = mb(path, create=create)
        except Exception as e:
            if isinstance(e, IOError):
                path += ' (maybe it\'s a different mailbox type?)'
            panic('Failed to open ', path, extb=E())
        return mb
# }}}

class Config: # {{{
    """Head and control class for configuration (--resource) data."""

    class Section:
        """Supertype for [SECTION] managers."""
        def __init__(self, typename):
            self._sectype = typename
            self._id = None

        def id_is_set(self):
            return not not self._id
        @property
        def id(self):
            return self._id if self._id is not None else '!id NOT SET!'
        @id.setter
        def id(self, id):
            """Protected - should only be called by subclasses."""
            self._id = id

        @property
        def full_id(self):
            return '[' + self._sectype + ':' + self.id +']'

        def set_attribute(self, name, val):
            return self.full_id + ': unknown directive: ' + name + '=' + val

        def close_config(self):
            """Section is closed.
            Return value is an error string or ``None``.
            """
            verb(self.full_id, ': close config ok')
            return None

        def finalize_config(self):
            """Configuration is entirely parsed, perform stuff which requires
            that all settings etc. are available.
            Return value is an error string or ``None``.
            """
            verb(self.full_id, ': finalize config ok')
            return None

    main = None
    """Configuration object - ini7d it via Config.setup_main()."""

    _sectre_ = re.compile(r'^\[\s*([-\w\.]+)\s*\](?:\s*#.*)?$')
    _kvre_ = re.compile(r'^(\S+)\s*=\s*(.*)$')

    @staticmethod
    def setup_main():
        Config.main = Config()

    def __init__(self):
        self._rcstack = []
        self._rc = None
        self._rclno = 0
        self._dict = {
            # [default]
            'mailbox-type': 'mbox',
            'ask-user': False,
            'fetch-hook': None, 'fetch-hook-largemail': None,
                'fetch-hook-treshold': 500000,
            'loop-check': 1200, 'loop-check-fast': 300,
                'loop-treshold': 1, 'loop-hook-newmail': None,
            # (Also) Command line
            'maildir': None,
            'ignore-errors': False,
            'keep-archives': False,
            '_start-value': None
        }
        self.__accountno = self.__targno = 0
        self.__finalizers = []
        self.__section = None

    def set_resource(self, rc):
        try:
            self._rc = os.path.abspath(os.path.expanduser(rc))
        except:
            panic('S-Postman: resource file path check failed', extb=E())
        log('S-Postman resource file: ', self._rc)
        # Pre-requisite conditions (don't care of atomicity)
        if not os.access(self._rc, os.R_OK):
            panic('S-Postman resource file ', self._rc, ' is not readable')
        if ((os.stat(self._rc).st_mode & 0o0777) & ~0o0600) != 0:
            panic('S-Postman resource file ', self._rc,
                  ' permissions failure!\n'
                  '  Only the active user shall be able to read (and write) it')

    def __getitem__(self, key):
        if not key in self._dict:
            panic('Config.__getitem__(): unknown key!')
        return self._dict[key]

    def __setitem__(self, key, value):
        if not key in self._dict:
            panic('Config.__setitem__(): unknown key!')
        self._dict[key] = value

    def generate_config(self, check_only=False):
        """Reads the resource file and should thus be called after all
        pre-setup has been finished, e.g. after the command line parsing has
        been completed.  It'll also create the mail directory as necessary,
        unless check_only is true.
        """
        if self._rc is None:
            self.set_resource(os.environ.get('S_POSTMAN_RC', POSTMAN_RC))

        self._parse()
        for f in self.__finalizers:
            em = f.finalize_config()
            if em is not None:
                error('Configuration: ', em)
        if self.__accountno == 0 and self.__targno == 0:
            panic(self._rc, ': incomplete: need [account]s and/or [TARGET]s')
        del self.__accountno, self.__targno, self.__finalizers, self.__section
        verb('Checking box branch limits:')
        Rule.check_loop_limits()
        verb('... box branch limits ok')

        if not check_only:
            self._maildir_check(check_only)

    def _verb(self, *args):
        if VERBOSE < 2:
            return
        if len(self._rcstack) > 0:
            verb('%INC:', self._rclno, ':', *args)
        else:
            verb('RC:', self._rclno, ':', *args)
    def _error(self, *args):
        error('In resource ', self._rc, ', line ', self._rclno, ':\n', *args)

    def _parse(self, file=None):
        if not file:
            file = self._rc
            verb('Reading ResourCe file ', file, ':')
        else:
            verb('%including ', file, ':')
        try:
            with open(file) as f:
                lines = f.readlines()
        except IOError as ioe:
            panic('Failed to open ', file, ' readonly', extb=E())
        self._rc = file
        self._rclno = 0

        cl = str() # (Continuation-Line)
        for l in lines:
            self._rclno += 1
            l = l.strip()
            if not len(l) or l.startswith('#'):
                if not len(cl):
                    continue
            elif l.endswith('\\'):
                cl += l[:-1]
                continue
            if len(cl):
                Config._match_line(self, cl + l)
                cl = str()
            else:
                Config._match_line(self, l)
        if len(cl):
            Config._match_line(self, cl)
        if self.__section is not None:
            cl = self.__section.close_config()
            if not not cl:
                self._error(cl)
            self.__section = None

    @staticmethod
    def _match_line(self, line):
        if line.startswith('%include '):
            self._include(line)
            return

        m = Config._sectre_.match(line)
        if m is not None:
            if self.__section is not None:
                em = self.__section.close_config()
                if not not em:
                    self._error(em)
                self.__section = None
            b = m.group(1).lower()
            if b == r'default':
                self._verb('Switching to [DEFAULT] section')
            else:
                self._verb('Creating [', b, ']')
                self._start_section(b)
            return

        m = Config._kvre_.match(line)
        if m is not None: # FIXME v = os.path.expandvars(v) ?? ALWAYS???
            k, v = m.group(1).lower(), m.group(2)
            if self.__section is None:
                self._verb('[DEFAULT]: K<', k, '> = V<', v, '>')
                if k == 'maildir':
                    if not self['maildir']:
                        self['maildir'] = v
                    else:
                        verb('maildir: overwritten by command line option')
                elif k == 'mailbox-type':
                    v = MailboxHelper.name_if_supported(v)
                    if not v:
                        self._error(v, ' is not a supported mailbox type')
                    self['mailbox-type'] = v
                elif k == 'ask-user':
                    self['ask-user'] = bool_check(v)
                # --loop
                elif k == 'loop-check':
                    try:
                        self['loop-check'] = int(v, 10)
                    except:
                        self._error(v, ': loop-check must be a number')
                elif k == 'loop-check-fast':
                    try:
                        self['loop-check-fast'] = int(v, 10)
                    except:
                        self._error(v, ': loop-check-fast must be a number')
                elif k == 'loop-treshold':
                    try:
                        self['loop-treshold'] = int(v, 10)
                    except:
                        self._error(v, ': loop-treshold')
                elif k in self._dict and not k.startswith('_'):
                    self[k] = v
                else:
                    self._error(k, ': unsupported configuration key')
            else:
                self._verb(self.__section.full_id, ': K<', k,
                            '> = V<', v, '>')
                v = self.__section.set_attribute(k, v)
                if v is not None:
                    self._error(v)
            return

        self._error('unsupported line content')

    def _start_section(self, section):
        finit = False
        if section == 'account':
            self.__accountno += 1
            self.__section = Account()
        elif section == 'box':
            self.__targno += 1
            self.__section = Box()
        #IMAP elif section == 'folder':
        #IMAP    self.__targno += 1
        #IMAP    self.__section = Folder()
        elif section == 'rule':
            self.__section = Rule()
            finit = True
        elif section == 'ruleset':
            self.__section = Ruleset.get_instance()
            finit = True
        elif section == 'hostfingerprints':
            self.__section = HostFingerPrints.get_instance()
        else:
            self._error('[', section, '] is of an unsupported type')

        if finit:
            self.__finalizers.append(self.__section)

    def _maildir_check(self, check_only):
        d = self['maildir']
        if not d:
            d = os.environ.get('S_POSTMAN_MAILDIR', POSTMAN_MAILDIR)
        d = os.path.expanduser(d)
        d = os.path.realpath(d)
        self['maildir'] = d
        log('S-Postman mail directory: ', d)
        if check_only:
            return

        if not os.path.isdir(d):
            log('Creating mail directory ', d)
            try:
                os.makedirs(d)
            except Exception as e:
                panic('Failed to create mail directory', extb=E())
        if not os.access(d, os.R_OK|os.W_OK):
            panic('S-Postman mail directory ', d, ' is not accessible')

    def _include(self, line):
        file = line[9:].strip() # (%include )
        if not len(file):
            self._error('%include without argument')
            return
        try:
            i = os.path.expanduser(file)
            if i == file:
                i = os.path.join(os.path.dirname(self._rc), file)
                i = os.path.abspath(i) # xxx first level only??
            file = i
        except:
            panic('%include failed on path expansion for ', file, extb=E())

        self._rcstack.append((self._rc, self._rclno))
        self._parse(file)
        self._rc, self._rclno = self._rcstack.pop()
        verb('%include finished, switching back to: ', self._rc)
# }}}

class HostFingerPrints(Config.Section): # {{{
    """FIXME"""
    _self_ = None

    @staticmethod
    def get_instance():
        if not HostFingerPrints._self_:
            HostFingerPrints._self_ = HostFingerPrints()
        return HostFingerPrints._self_

    def __init__(self):
        Config.Section.__init__(self, 'hostfingerprints')
        self._dict = {}
        self.id = ''

    def set_attribute(self, name, val):
        self._dict[name] = val
        return None

    def close_config(self):
        return None
# }}}

class Account(Config.Section): # {{{
    """[account]"""
    _accounts_ = []

    @staticmethod
    def get_accounts():
        return Account._accounts_

    @staticmethod
    def find_account(id):
        for a in Account._accounts_:
            if a.id == id:
                return a
        return None

    def __init__(self):
        Config.Section.__init__(self, 'account')
        self._account = None
        self._aliases = list()
        self._user = None
        self._pass = None
        self._send_type = None
        self._send_server = None
        self._fetch_type = None
        self._fetch_server = None
        self._ssl = False

    @property
    def account(self):      return self._account
    @property
    def aliases(self):      return self._alias

    @property
    def user(self):         return self._user
    @property
    def password(self):     return self._pass

    def credentials_ok(self):
        if not self._ssl or self._user and self._pass:
            return True
        return not not self._user

    def verify_credentials(self):
        """Get credentials from standart input as necessary, or ``panic()``."""
        if self.credentials_ok():
            return
        if not sys.stdin.isatty():
            panic('Account ' + self._account + ' requires credentials, '
                  'but standart input is not a terminal')

        while not self._user:
            l = user_line('[account] ' + self.id + ': username:')
            if user_confirm('<' + l + '> - is this correct [yY - else no]? '):
                self._user = l
                break

        if not self._pass:
            l = 'required' if self._ssl else 'optional'
            l = user_pass('[account] ' + self.id +
                          ': password (' + l + '): ')
            self._pass = l

    @property
    def send_type(self):    return self._send_type
    @property
    def send_server(self):  return self._send_server

    @property
    def fetch_type(self):   return self._fetch_type
    @property
    def fetch_server(self): return self._fetch_server

    def set_attribute(self, name, val):
        if name == 'id':
            if Account.find_account(id):
                return '[account] ' + id + ' already defined'
            self.id = val
        elif name == 'account':
            self._account = val
        elif name == 'alias':
            self._aliases.append(val)
        elif name == 'user':
            self._user = val
        elif name == 'pass':
            self._pass = val
        elif name == 'send-type':
            val = val.lower()
            if val != 'smtp' and val != 'smtps':
                return 'send-type \"' + val + '\" is not supported'
            self._send_type = val
            if val == 'smtps': self._ssl = True
        elif name == 'send-server':
            self._send_server = val
        elif name == 'fetch-type':
            val = val.lower()
            if val != 'pop' and val != 'pops':
                return 'fetch-type \"' + val + '\" is not supported'
            self._fetch_type = val
            if val == 'pops': self._ssl = True
        elif name == 'fetch-server':
            self._fetch_server = val
        else:
            return Config.Section.set_attribute(self, name, val)
        return None

    def close_config(self):
        if not self.id_is_set() or not self._account:
            n = self.id if self.id_is_set() else self._account
            if not n: n = 'UNKNOWN'
            return '[account] ' + n + ': \"id\" and \"account\" are required'
        if not Config.main['ask-user']:
            if not self._user:
                return ('[account] ' + self.id + ':\n  \"user\" is required '
                        'if ask-user is not set (in [default])')
            if self._ssl and not self._pass:
                return ('[account] ' + self.id + ': \"pass\" is required '
                        'for encrypted protocols '
                        '(if ask-user is not set in [default])')
        if self._send_server is None and self._fetch_server is None:
            return ('[account] ' + self.id + ': without '
                    'fetching and/or sending an account is useless')
        Account._accounts_.append(self)
        return Config.Section.close_config(self)

    def finalize_config(self):
        return None
# }}}

class Target(Config.Section): # {{{
    """The supertype of mail targets, i.e. ``[box]`` and ``[folder]``."""
    _targets_, _null_ = [], None

    @staticmethod
    def find_target(name):
        """Find a target which' ``id`` property matches ``name``, or None."""
        if name.lower() == 'null':
            if not Target._null_:
                Target._null_ = Null()
            return Target._null_
        for t in Target._targets_:
            if t.id == name:
                return t
        return None

    @staticmethod
    def get_last_defined():
        """Get last defined target or None if no target exists at all."""
        return Target._targets_[-1] if len(Target._targets_) else None

    @staticmethod
    @atexit.register
    def __close():
        """It's ensured that close() is called for subclasses even on error."""
        if not len(Target._targets_):
            return
        for t in Target._targets_:
            try:
                t.close()
            except Exception as e:
                error('Failed to close target ', t.id, logonly=True, extb=E())

    def __init__(self, tname):
        Config.Section.__init__(self, tname)
        self._saved_any = False

    def close_config(self):
        self._targets_.append(self)
        return Config.Section.close_config(self)

    def finalize_config(self):
        return None

    def close(self):
        """Auto-called subclass hook to ensure proper cleanup on exit.
        This may raise an exception - Target.__close() will handle that.
        """
        pass

    def has_saved_any(self):    return self._saved_any
    def set_saved_any(self):    self._saved_any = True

    def save_ticket(self, ticket, ignore_errors=False):
        """Save a ticket in that target.  Return a boolean."""
        return False
# }}}

class Null(Target): # {{{
    """Single instance NULL target, created by Target as necessary."""
    def __init__(self):
        Target.__init__(self, '[NULL]')
        self.id = 'null'
    def save_ticket(self, ticket, ignore_errors=False):
        log('@ Sinking ticket ', ticket.id, ' in null')
        self.set_saved_any()
        return True
# }}}

class Box(Target): # {{{
    """[box]"""
    def __init__(self):
        Target.__init__(self, 'box')
        self._path = None
        self._type = Config.main['mailbox-type']
        self._mailbox = None

    @property
    def path(self): return self._path
    @property
    def type(self): return self._type

    def set_attribute(self, name, val):
        if name == 'id':
            if val.lower() == 'null':
                return 'NULL cannot be used for [box] - it\'s a builtin'
            if not not Target.find_target(val):
                return '[box] ' + val + ' already defined'
            self.id = val
        elif name == 'path':
            try:
                self._path = os.path.abspath(os.path.expanduser(val))
            except:
                return 'could not expand path ' + val
        elif name == 'type':
            self._type = MailboxHelper.name_if_supported(val)
            if not self._type:
                return val + ' is not a supported mailbox type'
        else:
            return Target.set_attribute(self, name, val)
        return None

    def close_config(self):
        if not self.id_is_set():
            n = self._path
            if n is None: n = self.full_id
            return '[box] ' + n + ': requires an \"id\"'
        if self._path is None:
            self._path = self.id
        return Target.close_config(self)

    def close(self):
        if self._mailbox:
            log(self.id, ': closing mailbox')# FIXME
            self._mailbox.close()

    def save_ticket(self, ticket, ignore_errors=False):
        log('@ Saving ticket ', ticket.id, ' in ', self.id)
        efun = panic
        if ignore_errors:
            efun = error

        mb = self._mailbox
        if not mb:
            mb = os.path.join(Config.main['maildir'], self._path)
            mb = MailboxHelper.open(mb, type=self._type, create=True)
            self._mailbox = mb
            try:
                mb.lock()
            except Exception as e:
                efun('Could not gain mailbox lock', extb=E())
                # FIXME RETURN

        try:
            mb.add(ticket.message())
            mb.flush()
        except Exception as e:
            efun('[box] ', self.id, ': message-add failed.\n',
                 'FIXME about fetch modus which saves the stuff somewhere',
                extb=E())
        finally:
            mb.unlock()

        self.set_saved_any()
        return True # FIXME
# }}}

# Rule(set)-Verbose, for easier grep+replace
_rv = verb

class Ruleset(Config.Section): # {{{
    """[ruleset]"""
    class Match:
        """Returned by ``Ruleset.get_instance().match_ticket()``.
        ``splitter`` is a callable reference or ``None`` and ``targets``
        is a list of ``Target`` subclasses.
        """
        def __init__(self, is_match, splitter, targets):
            self._is_match = is_match
            self._splitter = splitter
            self._targets = targets[:]

        @property
        def is_match(self): return self._is_match
        @property
        def splitter(self): return self._splitter
        @property
        def targets(self):  return self._targets

    _self_ = None

    @staticmethod
    def get_instance():
        if not Ruleset._self_:
            Ruleset._self_ = Ruleset()
        return Ruleset._self_

    def __init__(self):
        Config.Section.__init__(self, 'ruleset')
        self.id = ''
        self._rules = []
        # Pre-check: does header field has any associated rule at all?
        self._mail_fields = set()

    def set_attribute(self, name, val):
        targets = []
        for t in val.split(','):
            t = t.lower()
            if not len(t):
                continue # XXX error case empty sequence?  Log it?
            targets.append(t.strip())
        # Special case the special NULL rule with NULL sink right here, so
        # 'NULL = ' is possible..
        if not len(targets) and name.lower() != 'null':
            return '[ruleset]: rule ' + name + ': empty target specification'
        self._rules.append((name, targets))
        return None

    def close_config(self):
        return None

    def finalize_config(self):
        if not len(self._rules):
            return '[ruleset] may not be empty'
        for i in range(len(self._rules)):
            (name, targets) = self._rules[i]

            rule = Rule.find_rule(name)
            if not rule:
                return '[ruleset]: [rule] ' + name + ' was never defined'

            for j in range(len(targets)):
                tn = targets[j]
                t = Target.find_target(tn)
                if not t:
                    return ('[ruleset]: [rule] ' + name + ': target ' + tn +
                            ' was never defined')
                targets[j] = t

            self._rules[i] = (rule, targets)
        return Config.Section.finalize_config(self)

    def add_mail_field(self, field):
        self._mail_fields.add(field)

    def match_ticket(self, ticket):
        # Create intersection of self._mail_fields and ticket fields
        fields = set()
        for f in ticket.fields():
            if f in self._mail_fields:
                fields.add(f)
        if _rv == verb:
            _rv('    + May match headers: ', [f for f in iter(fields)])

        if len(fields):
            verb('    @ Searching [rule]s')
            for (rule, targets) in self._rules:
                m = rule.match_ticket(1, ticket, fields, targets)
                if m:
                    return m
        log('    @ Ticket doesn\'t match any [rule]s - using last target')
        return Ruleset.Match(False, None, [Target.get_last_defined()])
# }}}

class Rule(Config.Section): # {{{
    """[rule]"""
    class _Instr:
        PASS, HIT, BRANCH, ANY, SPLITTER, PLUG,PLUG_ADD,PLUG_SUB = \
            1, 2, 3, 4, 5, 6,7,8
        MATCH_COMM_MAX = 4

        def __init__(self, type, keys=None, value=None, value_re=None, log=''):
            """A parsed instruction.

            If ``type=PLUG*``, then ``value`` *may* be set to a target.
            If ``type=SPLITTER``, then ``keys`` is it's name and ``value``
            is a function reference, unless it's a reset action (all ``None``).

            ``PASS`` and ``HIT`` use the full set of arguments: ``keys`` is
            a ``set`` of (mail header) fields to which this instruction
            applies, ``value`` is the cleaned up regex as a normal string,
            and ``value_re`` the same, but as a compiled ``re`` object.
            ``log`` is used for verbose logging the line and thus consists
            of the entire instruction (including type) as a string.

            ``BRANCH`` is equal to ``PASS``, but ``value`` refers to the ID of
            the rule to branch to instead.
            It is later updated to the rule instance due to performance issues.
            """
            self._type = type
            self._keys = keys
            self._val = value
            self._val_re = value_re
            self._log = log

        @property
        def type(self):         return self._type
        @property
        def keys(self):         return self._keys
        @property
        def value(self):        return self._val
        @value.setter
        def value(self, newv):  self._val = newv
        @property
        def value_re(self):     return self._val_re
        @property
        def log(self):          return self._log

    _rules_, _null_ = {}, None
    SEEN_NONE, SEEN_PASS, SEEN_UNCOND_BRANCH, SEEN_ANY = 0, 1<<0, 1<<1, 1<<2

    @staticmethod
    def find_rule(id):
        if id == 'null':
            if not Rule._null_:
                r = Rule()
                r.id = 'null'
                r._instr6s.append(r._Instr(r._Instr.PLUG,
                                           value=Target.find_target('null')))
                r._instr6s.append(r._Instr(r._Instr.ANY))
                r._seen = r.SEEN_UNCOND_BRANCH | r.SEEN_ANY
                del r.__real_instr6s
                Rule._null_ = r
            return Rule._null_
        return Rule._rules_.get(id)

    @staticmethod
    def check_loop_limits():
        def _follow_branch(id, cnt, rule):
            cnt += 1
            if cnt > 99:
                panic('[rule] ' + id + ': BRANCH exceeds 99 levels '
                      'of recursion')
            for ins in rule._instr6s:
                if ins.type != Rule._Instr.BRANCH:
                    continue
                _follow_branch(id, cnt, ins.value)

        for r in Rule._rules_.values():
            _follow_branch(r.id, 0, r)

    def __init__(self):
        Config.Section.__init__(self, 'rule')
        self._instr6s = []
        self._fields = set()
        self._seen = self.SEEN_NONE
        self.__real_instr6s = 0

    def set_attribute(self, name, val): # {{{
        if name == 'id':
            self.id = val
            return None

        if name == '$':
            if not self.__real_instr6s:
                return ('ANY is valid only after at least one PASS, HIT or '
                        'BRANCH')
            self._instr6s.append(self._Instr(self._Instr.ANY))
            self._seen |= self.SEEN_ANY
            verb('  ANY')
            return None

        if name.startswith(':'):
            if self._seen & self.SEEN_ANY:
                return 'Commands cannot be used after ANY'
            t = self._Instr.PLUG
            if len(name) > 1:
                if name[1] == '+':
                    name = 'PLUG-ADD'
                    t = self._Instr.PLUG_ADD
                elif name[1] == '-':
                    name = 'PLUG-SUB'
                    t = self._Instr.PLUG_SUB
                else:
                    return 'Unknown PLUG command specified'
            else:
                name = 'PLUG'

            val = val.strip()
            if not len(val):
                if t != self._Instr.PLUG:
                    return name + ' requires box specifications'
                self._instr6s.append(self._Instr(self._Instr.PLUG))
                verb('  PLUG: use default target list')
                return None

            val = Target.find_target(val)
            if val is None:
                return name + ' command specifies non-existent target'
            if val.id == 'null' and t != self._Instr.PLUG:
                return name + ' cannot be used with NULL data sink'
            self._instr6s.append(self._Instr(t, value=val))
            verb('  ', name, ': ', val.id)
            return None

        if name == '*':
            if self._seen & self.SEEN_ANY:
                return 'Commands cannot be used after ANY'
            val = val.strip()
            if len(val):
                (name, val) = Splitter.create_if_supported(val)
                if val is None:
                    return 'SPLITTER command specifies unknown splitter type'
                self._instr6s.append(self._Instr(self._Instr.SPLITTER,
                                                 keys=name, value=val))
                verb('  SPLITTER: ', name)
            else:
                self._instr6s.append(self._Instr(self._Instr.SPLITTER))
                verb('  SPLITTER: none')
            return None

        if name[0] != '?' and name[0] != '!' and name[0] != '@':
            return Config.Section.set_attribute(self, name, val)
        # PASS, HIT, BRANCH
        if self._seen & self.SEEN_ANY:
            return 'Commands cannot be used after ANY'

        self.__real_instr6s += 1
        valre = val = val.strip()
        is_branch = None

        if name.startswith('@'):
            m = re.match(r'^@([^@]+)(?:@(.+))?$', name)
            if not m:
                return 'illegal BRANCH specification: ' + name
            val = m.group(1)
            name = '#'
            if m.group(2):
                name += m.group(2)
                is_branch = valre
            else:
                if len(valre):
                    return 'no header fields but a regular expression'
                is_branch = '(empty)'
                if not self._seen & self.SEEN_ANY:
                    self._seen |= self.SEEN_UNCOND_BRANCH

        if len(valre):
            try:
                valre = re.compile(valre)
            except Exception as e:
                return ('illegal regular expression: ' + valre + '\n  ' +
                        str(e))
        elif not is_branch:
            return 'empty regular expression value'

        fields = set()
        for t in name[1:].split(','):
            t = t.lower().strip()
            if not len(t):
                continue # XXX error case empty sequence?  Log it?
            if re.match(r's(?:ub(?:ject)?)?', t):
                fields.add('subject')
            elif re.match(r'f(?:rom)?', t):
                fields.add('from')
            elif re.match(r'to?', t):
                fields.add('to')
            elif re.match(r'cc?', t):
                fields.add('cc')
            elif re.match(r'b(?:c(?:c)?)?', t):
                fields.add('bcc')
            else:
                fields.add(t)

        if not len(fields):
            if not is_branch and isinstance(valre, str):
                return 'Command does not specify any header field names'
            name_str = 'no conditions'
        else:
            name_str = ''
            for f in iter(fields):
                Ruleset.get_instance().add_mail_field(f)
                self._fields.add(f)
                name_str += f + ','
            name_str = name_str[:-1]

        if name.startswith('?'):
            self._seen |= self.SEEN_PASS
            r = self._Instr.PASS
            name = 'PASS' + '<' + name_str + ' = ' + val + '>'
        elif name.startswith('!'):
            r = self._Instr.HIT
            name = 'HIT' + '<' + name_str + ' = ' + val + '>'
        else:
            r = self._Instr.BRANCH
            name = ('BRANCH' + '<' + val + '@' + name_str + ' = ' +
                    is_branch + '>')
        self._instr6s.append(self._Instr(r, keys=fields,
                                         value=val, value_re=valre, log=name))
        verb('  ', name)
        return None
    # }}}

    def close_config(self):
        if not self.id_is_set():
            return '[rule] ' + n + ': requires an \"id\"'
        if not self.__real_instr6s:
            return '[rule] without any PASS, HIT or BRANCH commands'
        del self.__real_instr6s

        t = self._instr6s[-1].type
        if t == self._Instr.PASS:
            return '[rule]: a final PASS command (?) is useless'
        if t == self._Instr.SPLITTER:
            return '[rule]: a final SPLITTER command (+) is useless'
        if (t == self._Instr.PLUG or t == self._Instr.PLUG_ADD or
            t == self._Instr.PLUG_SUB):
            return '[rule]: a final PLUG command (+) is useless'

        Rule._rules_[self.id] = self
        return Config.Section.close_config(self)

    def finalize_config(self):
        for i in self._instr6s:
            if i.type == self._Instr.BRANCH:
                r = self.find_rule(i.value)
                if not r:
                    return '[rule]: BRANCH ' + i.value + ' never defined'
                i.value = r
        return Config.Section.finalize_config(self)

    def match_ticket(self, lvl, ticket, fields, base_targets): # {{{
        dash = '-' * lvl
        plus = '+' * lvl
        at = '@' * lvl
        _rv('    ', plus, ' [rule] ', self.id)

        # Intersection of fields of self & Ruleset.match_ticket()'s global one
        fieldset = fields & self._fields
        if not len(fieldset):
            if not self._seen & self.SEEN_UNCOND_BRANCH:
                _rv('      ', dash, ' No match intersection')
                return None
            _rv('      ', dash, ' No match intersection, '
                'but unconditional branch')
        elif _rv == verb:
            _rv('      ', plus, ' Match intersection: ', fieldset)

        targets = base_targets[:]
        splitter = None
        fld = None

        for ins in self._instr6s:
            # Match-commands first
            if ins.type <= ins.MATCH_COMM_MAX:
                if ins.type == ins.ANY:
                    log('    ', at, ' ANY match [rule] ', self.id)
                    return Ruleset.Match(True, splitter, targets)

                # For all fields in intersection, check wether field is in
                # this command; if so, check wether any of the tickets'
                # bodies of that field matches the commands value regex
                _rv('      ', dash, ' Command ', ins.log)
                for fld in fieldset:
                    if fld not in ins.keys:
                        continue
                    for body in ticket.bodies_for_field(fld):
                        if ins.value_re.search(body) is None:
                            continue
                        if ins.type == ins.HIT:
                            log('    ', at, ' HIT match [rule] ', self.id,
                                ', command: ', ins.log)
                            return Ruleset.Match(True, splitter, targets)
                        if ins.type == ins.BRANCH:
                            m = ins.value.match_ticket(lvl+1, ticket, fields,
                                                       targets)
                            if m:
                                return m
                        break
                    else:
                        continue
                    _rv('        ', dash, ' PASS match ok, next command')
                    break
                else:
                    if ins.type == ins.PASS:
                        _rv('        ', dash, ' PASS failed, exit rule')
                        return None
                    if ins.type == ins.BRANCH:
                        # Unconditional branch?
                        if not len(ins.keys):
                            m = ins.value.match_ticket(lvl+1, ticket, fields,
                                                       targets)
                            if m:
                                return m
                        _rv('        ', dash, ' BRANCH miss, next command')
                    else:
                        _rv('        ', dash, ' HIT failed, next command')
                continue

            # Environmental commands afterwards (no goto here ;-()
            if ins.type == ins.PLUG:
                if not ins.value:
                    targets = base_targets[:]
                else:
                    targets = [ins.value]
            elif ins.type == ins.PLUG_ADD:
                targets.append(ins.value)
            elif ins.type == ins.PLUG_SUB:
                try:
                    targets.remove(ins.value)
                except:
                    error('PLUG-SUB: ', ins.value.id, ' is not plugged at all',
                          logonly=True)
            elif ins.type == ins.SPLITTER:
                splitter = ins.value
            #continue

        _rv('      ', plus, ' Out of commands')
        return None
        # }}}
# }}}

class Ticket: # {{{
    """Don't pass around email.message stuff, but encapsulate in a ticket."""
    _ticket_count_ = 0

    @staticmethod
    def process_message(msg):
        ticket = Ticket(msg)
        rsm = Ruleset.get_instance().match_ticket(ticket)
        splitter = rsm.splitter
        if rsm.is_match and splitter and not Config.main['keep-archives']:
            log('    @ Treating ticket ', ticket._id, ' as archive, splitting')
            splitter = splitter(msg)
            if splitter:
                return [Ticket(m, _targets=rsm.targets) for m in splitter]
        ticket._targets = rsm.targets
        return [ticket]

    def __init__(self, msg, _targets=None):
        Ticket._ticket_count_ += 1
        self._id = Ticket._ticket_count_
        self._msg = msg
        self._targets = _targets
        self._fields = None # {field:[bodies],..}
        self._bewitch_msg()
        log('  @ Creating ticket number ', self._id, ':')

        f = msg['From']
        if not f:       f = 'NO From:'
        f = str(f)
        t = msg['To']
        if not t:       t = 'NO To:'
        t = str(t)
        s = msg['Subject']
        if not s:       s = '<No Subject>'
        s = str(s)
        log('    From:     ', f)
        log('    To:       ', t)
        log('    Subject:  ', s)
        self._syno = '#{:.14}#{:.14}#{:.20}#'.format(f, t, s)
        log('    Synopsis: ', self._syno)

    @property
    def id(self):
        """Unique creation sequence number."""
        return self._id

    @property
    def synopsis(self):
        """Single line abbreviation showing most important keys."""
        return self._syno

    @property
    def targets(self):
        """A list of all matching targets or ``None``."""
        return self._targets

    def fields(self):
        """The ``list()`` of message header fields."""
        if self._fields is None:
            self._fields = {}
            for k in self._msg.keys():
                k = k.lower()
                self._fields[k] = None
        return self._fields.keys()

    def bodies_for_field(self, field):
        """A ``list()`` of all field bodies for the header ``field``."""
        if self._fields is None:
            self.fields()
        field = field.lower()
        bl = self._fields[field]
        if bl is None:
            bl = list(str(b) for b in self._msg.get_all(field, []))
            self._fields[field] = bl
        return bl

    def message(self):
        """The ``email.message.Message`` is only for targets."""
        return self._msg

    def _bewitch_msg(self):
        """Handle Python 3.2.0/3.3a0 issue 11401 email/message.py error"""
        if (sys.hexversion > 0x030300A0 or
            sys.hexversion < 0x03030000 and sys.hexversion > 0x030200F0):
            return
        # >>DAVID PATCH
        # This is not part of the normal version
        alln = self._msg.items()[:] # In fact -> ensure all are header.Header
        # If any converted (str->Header) header names exist ...
        if len(alln):
                # Delete *all* occurrences of h (doesn't throw)
                for (n, b) in alln:
                    del self._msg[n]
                # And append in order
                for (n, b) in alln:
                    self._msg[n] = email.header.make_header(email.header.decode_header(b))
        # <<DAVID PATCH
        for f in self._msg:
            had_repl = False
            ab = self._msg.get_all(f)
            for i in range(len(ab)):
                if not len(str(ab[i])):
                    ab[i] = ' '
                    had_repl = True
            if had_repl:
                del self._msg[f]
                for b in ab:
                    self._msg[f] = b
# }}}

class Dispatch: # {{{
    _dispatch_boxes_ = []

    @staticmethod
    def add_box(box):
        box = os.path.expanduser(box)
        # It's path[:type]
        m = re.match(r'(.+):(.+)', box)
        if m is not None:
            t = m.group(2).lower()
            n = MailboxHelper.name_if_supported(t)
            if not n:
                panic('--dispatch=', box, ': ', t,
                      ' is an unsupported mailbox type')
            t = n
            box = m.group(1)
            verb('--dispatch: box ', box, ' is of type ', t)
        else:
            t = 'mbox'

        # First simple non-atomic check
        if not os.access(box, os.R_OK):
            panic('--dispatch: box ', box, ' is not readable')
        Dispatch._dispatch_boxes_.append([box, t])

    def __init__(self):
        self._tickets = []
        self._walk()

    def _walk(self):
        drop_count = 0
        if Config.main['_start-value']:
            try:
                drop_count = int(Config.main['_start-value'], 10)
            except:
                panic('--dispatch: --start= argument must be a number')

        verb('--dispatch: iterating over input boxes')
        for b, t in self._dispatch_boxes_:
            mb = MailboxHelper.open(b, type=t, create=False)
            try:
                log('* Box has ', len(mb),
                    ' messages, filtering through rulesets')
                cnt = 0
                for k in mb.iterkeys():
                    cnt += 1
                    log('({!s:>04})'.format(cnt))
                    msg = mb.get_message(k)
                    self._tickets.extend(Ticket.process_message(msg))
            except Exception as e:
                error('failed to handle box ', b, extb=E())
            finally:
                try:
                   mb.close()
                except:
                    pass

        verb('--dispatch: iterating over processed tickets')
        errors = tcnt = bcnt = 0
        for ticket in self._tickets:
            if ticket.id < drop_count:
                verb('@ Dropping ticket ', ticket.id, ': before --start')
                continue
            tcnt += 1
            for target in ticket.targets:
                if not target.has_saved_any():
                    bcnt += 1
                if not target.save_ticket(ticket):
                    errors += 1
        tcnt = '1 ticket to' if tcnt == 1 else str(tcnt) + ' tickets to'
        bcnt = '1 mailbox' if bcnt == 1 else str(bcnt) + ' mailboxes'
        print('Dispatched', tcnt, bcnt)
        if errors:
            error(errors, ' errors occurred while saving mails') # FIXME
# }}}

class Fetch: # {{{
    _accounts_ = [] # empty == fetch all

    @staticmethod
    def add_account(account):
        Fetch._accounts_.append(account)

    def __init__(self):
        print('FETCH is not yet implemented')
        #FIXME: only SSL if any of the fetch woth pops
        try:
            from poplib import POP3_SSL
        except:
            panic('SSL support is required for POP3 mail fetching')

        # Get accounts to fetch
        if not len(Fetch._accounts_):
            Fetch._accounts_ = Account.get_accounts()
        else:
            x = Fetch._accounts_
            Fetch._accounts_ = []
            for id in x:
                a = Account.find_account(id)
                if not a:
                    error('--fetch specified illegal account: ' + id)
                else:
                    Fetch._accounts_.append(a)
            if not len(Fetch._accounts_):
                panic('--fetch requires at least one account')

## FIXME fetch mode REQUIRES a pipe hook so that spamassassin plus can be used:
#       in fact offer ONLY special spamc(1) -c (fast part of that suite) support
#       (otherwise the entire message must be piped through and rebuild from
#       scratch!!)
# fetch-hook
# fetch-hook-largefile
# fetch-hook-large-treshold
# CHECK if pop,imap offer message-size in favour, otherwise files must be saved
# on disk before   CAT | HOOK | PROCESSED_FILe, not RECV|HOOK|FILE

        for a in Fetch._accounts_:
            a.verify_credentials()

        for a in Fetch._accounts_:
            print("FETCHING ACCOUNT", a.id)

        # email.feedparser.FeedParser
# }}}

class FetchLoop: # {{{
    def __init__(self):
        print('FETCHLOOP is not yet implemented')

        try:
            signal.signal(signal.SIGUSR1, FetchLoop._sigusr1)
        except:
            print('Wakeup via \'kill -USR1 xy\' is not supported - you win')


    @staticmethod
    def _sigusr1(signo, frame):
        return
# }}}

def command_line(): # {{{
    global VERBOSE
    def version():
        print(__version__, __copyright__.strip(), BUGMAIL, sep='\n')
    def doc():
        print(RST, USAGE, __doc__)
        version()
    def help():
        version()
        print('\n', USAGE.strip(), sep='')

    try:
        opts, args = getopt.getopt(sys.argv[1:],
                'hdvVm:r:ICF:ALD:ks:',
                ['help', 'doc', 'version', 'verbose',
                'maildir=', 'md=', 'resource=', 'rc=',
                'ignore-errors',
                'check-resource',
                'fetch=', 'fetchall', 'fetchloop',
                'dispatch=',
                'keep-archives', 'start='])
    except getopt.GetoptError as e:
        panic('Command line parsing failed: ', str(e))

    xclass = None
    for o, a in opts:
        if o in ('-h', '--help'):
            help()
            sys.exit(0)
        elif o in ('-d', '--doc'):
            doc()
            sys.exit(0)
        elif o in ('-v', '--version'):
            version()
            sys.exit(0)
        elif o in ('-V', '--verbose'):
            VERBOSE += 1
        elif o in ('-m', '--maildir', '--md'):
            Config.main['maildir'] = a
        elif o in ('-r', '--resource', '--rc'):
            Config.main.set_resource(a)
        elif o in ('-I', '--ignore-errors'):
            Config.main['ignore-errors'] = True
        elif o in ('-C', '--check-resource'):
            if xclass is not None and xclass is not Fetch:
                panic('--check-resource: MODE already set')
            class DummyClass:
                def __init__(self):
                    print('Healthy')
            xclass = DummyClass
        elif o in ('-F', '--fetch'):
            if xclass is not None and xclass is not Fetch:
                panic('--fetch: MODE already set')
            xclass = Fetch
            if not len(a):
                panic('empty --fetch argument')
            Fetch.add_account(a)
        elif o in ('-A', '--fetchall'):
            if xclass is not None:
                panic('--fetchall: MODE already set')
            xclass = o
        elif o in ('-L', '--fetchloop'):
            if xclass is not None:
                panic('--fetchloop: MODE already set')
            xclass = FetchLoop
        elif o in ('-D', '--dispatch'):
            if xclass is not None and xclass is not Dispatch:
                panic('--dispatch: MODE already set')
            xclass = Dispatch
            if not len(a):
                panic('empty --dispatch argument')
            Dispatch.add_box(a)
        elif o in ('-k', '--keep-archives'):
            Config.main['keep-archives'] = True
        elif o in ('-s', '--start'):
            Config.main['_start-value'] = a

    if xclass is None or isinstance(xclass, str):
        xclass = Fetch
    return xclass
# }}}

def main():
    try:
        Config.setup_main()
        xclass = command_line()
        Config.main.generate_config(check_only=(xclass.__name__=='DummyClass'))
        xclass = xclass()
    except Exception as e:
        panic('You\'ve found a bug - unhandled exception occurred!', extb=E())
    return ESTAT

if __name__ == '__main__':
    sys.exit(main())

# vim:set fenc=utf-8 filetype=python syntax=python ts=4 sts=4 sw=4 et tw=79:
